/*
	(C) 1995-2001 AROS - The Amiga Research OS
	$Id: closewindow.c,v 1.17 2021/03/28 02:52:30 jacadcaps Exp $

	Desc:
	Lang: english
*/
#define AROS_ALMOST_COMPATIBLE
#include <proto/exec.h>
#include <proto/graphics.h>
#include <proto/layers.h>
#include <proto/timer.h>
#include "intuition_intern.h"
#include "inputhandler.h"
#include "inputhandler_actions.h"
#include "inputhandler_support.h"
#include "segtracker.h"
#include "mosmisc.h"
#include "screenmanager.h"
#include "layermanager.h"
#include <intuition/gadgetclass.h>

#ifndef DEBUG_CloseWindow
#   define DEBUG_CloseWindow 1
#endif
#undef DEBUG
#if DEBUG_CloseWindow
#   define DEBUG 1
#endif
#	include <aros/debug.h>

/******************************************************************************/

#define MUST_UNLOCK_SCREEN(window,screen) (((IS(screen)->pubScrNode != NULL) && \
(window->MoreFlags & WMFLG_DO_UNLOCKPUBSCREEN)) ? TRUE : FALSE)

struct CloseWindowActionMsg
{
	struct IntuiActionMsg msg;
	struct Window *window;
};

VOID int_closewindow(struct CloseWindowActionMsg *msg,
					 struct IntuitionBase *IntuitionBase);

/*****************************************************************************

	NAME */
#include <proto/intuition.h>

AROS_LH1(void, CloseWindow,

		 /*  SYNOPSIS */
		 AROS_LHA(struct Window *, window, A0),

		 /*  LOCATION */
		 struct IntuitionBase *, IntuitionBase, 12, Intuition)

/*  FUNCTION
	Closes a window. Depending on the display, this might not happen
	at the time when this function returns, but you must not use
	the window pointer after this function has been called.

	INPUTS
	window - The window to close

	RESULT
	None.

	NOTES
	The window might not have been disappeared when this function returns.

	EXAMPLE

	BUGS

	SEE ALSO
	OpenWindow(), OpenWindowTags()

	INTERNALS

	HISTORY
	29-10-95    digulla automatically created from
			    intuition_lib.fd and clib/intuition_protos.h

*****************************************************************************/
{
	AROS_LIBFUNC_INIT
	AROS_LIBBASE_EXT_DECL(struct IntuitionBase *,IntuitionBase)

	struct CloseWindowActionMsg	msg;
	struct MsgPort		*userport;
	struct IntuiMessage *messagecache;
	struct Screen		*screen;
	BOOL			do_unlockscreen;

	DEBUG_CLOSEWINDOW(dprintf("CloseWindow: Window 0x%lx\n", window));

	D(bug("CloseWindow (%p)\n", window));

	if ( window == NULL )
	{
		ReturnVoid ("CloseWindow");
	}

	if (IsDuringSelfResize(IntuitionBase))
	{
		dprintf("Major bug: calling CloseWindow() from within MUIM_Draw task %lx\n",(ULONG)FindTask(0));
		return;
	}

	// special case
	if (IW(window)->specialflags & SPFLAG_DONOTCLOSE) return;

	screen = window->WScreen;
	do_unlockscreen = MUST_UNLOCK_SCREEN(window, screen);

#ifndef __MORPHOS__
	/* We take a very simple approach to avoid race conditions with the
	   intuition input handler running one input.device 's task:
	   We just send it a msg about closing the window
	*/
	if (HAS_CHILDREN(window))
	{
		struct Window * cw = window->firstchild;
		/*
		 * First close all its children, we could also return
		 * a failure value which newer apps that use child windows
		 * have to look at.
		 */
		while (cw)
		{
			struct Window * _cw;
			_cw = cw->nextchild;
			CloseWindow(cw);
			cw = _cw;
		}

	}

	if (IS_CHILD(window))
	{
		/*
		 * Unlink the window from its parent or
		 * out of the list of child windows.
		 */
		if (window->parent->firstchild == window)
			window->parent->firstchild = window->nextchild;
		else
			window->prevchild->nextchild = window->nextchild;

		if (window->nextchild)
			window->nextchild->prevchild = window->prevchild;

	}
#endif

	/* We must save this here, because after we have returned from
	   the Wait() the window is gone  */
	userport = window->UserPort;
	messagecache = IW(window)->messagecache; IW(window)->messagecache = NULL;

	if ((((struct IntWindow *)window)->specialflags & SPFLAG_USERPORT) && userport)
	{
		/* We must strip messages sent to this window!!! */
		struct IntuiMessage *im;

		Forbid();

		im = (struct IntuiMessage *)GetTail(&window->UserPort->mp_MsgList);

		while (im)
		{
			struct IntuiMessage *replyim = NULL;

			if (im->IDCMPWindow == window) replyim = im;

			im = (struct IntuiMessage *)GetPred(&im->ExecMessage.mn_Node);

			if (replyim)
			{
				REMOVE(&replyim->ExecMessage.mn_Node);
				replyim->IDCMPWindow = NULL;
				ReplyMsg((struct Message *)replyim);
			}
		}

		Permit();

		userport = NULL;
	}

	DEBUG_CLOSEWINDOW(dprintf("CloseWindow: Userport 0x%lx\n", userport));

	DEBUG_CLOSEWINDOW(dprintf("CloseWindow: DoSyncAction\n"));

#ifdef USEWINDOWLOCK
	/* wait until other open/close && move/size && menus actions are finished */
	if (!(FindTask(0) == ((struct IIHData *)IBase->InputHandler->is_Data)->InputDeviceTask))
	{
		LOCKWINDOW;
	}
#endif

	msg.window = window;
	DoSyncAction((APTR)int_closewindow, &msg.msg, IntuitionBase);

#ifdef USEWINDOWLOCK
	if (!(FindTask(0) == ((struct IIHData *)IBase->InputHandler->is_Data)->InputDeviceTask))
	{
		UNLOCKWINDOW;
	}
#endif

	if (do_unlockscreen)
	{
		UnlockPubScreen(NULL, screen);
	}

	/* As of now intuition has removed us from th list of
	   windows, and we will recieve no more messages
	*/

	if (messagecache)
	{
		messagecache->IDCMPWindow = 0;//zero or we'll trash mem in inputhandler!
		messagecache->Code = 0;
		messagecache->Qualifier = 0;
		ReplyMsg(&messagecache->ExecMessage);
	}

	if (userport)
	{
		struct IntuiMessage *im;

		DEBUG_CLOSEWINDOW(dprintf("CloseWindow: Reply UserPort Msgs\n"));
		while ((im = (struct IntuiMessage *) GetMsg (userport)))
		{
			im->IDCMPWindow = 0;
			im->Code = 0;
			im->Qualifier = 0;
			ReplyMsg ((struct Message *)im);
		}

		/* Delete message port */
		DeleteMsgPort (userport);
	}

	DEBUG_CLOSEWINDOW(dprintf("CloseWindow: done\n"));
	ReturnVoid ("CloseWindow");
	AROS_LIBFUNC_EXIT
} /* CloseWindow */

/******************************************************************************/

/* This is called from the intuition input handler */
VOID int_closewindow(struct CloseWindowActionMsg *msg,
					 struct IntuitionBase *IntuitionBase)
{
	/* Free everything except the applications messageport */
	ULONG		lock = 0;

	struct Window	*window, *win2;
	struct Screen	*screen;
	struct IIHData	*iihd;

	D(bug("CloseWindow (%p)\n", window));

	window = msg->window;

	iihd = (struct IIHData *)IBase->InputHandler->is_Data;

	/* if there is an active gadget in the window being closed, then make
	   it inactive */
	if (iihd->ActiveGadget &&
		!IS_SCREEN_GADGET(iihd->ActiveGadget) &&
		(iihd->GadgetInfo.gi_Window == window))
	{
		struct Gadget *gadget = iihd->ActiveGadget;

		switch(gadget->GadgetType & GTYP_GTYPEMASK)
		{
		case GTYP_CUSTOMGADGET:
			{
				struct gpGoInactive gpgi;

				gpgi.MethodID = GM_GOINACTIVE;
				gpgi.gpgi_GInfo = &iihd->GadgetInfo;
				gpgi.gpgi_Abort = 1;

				Locked_DoMethodA(window, gadget, (Msg)&gpgi, IntuitionBase);

				if (iihd->ActiveSysGadget)
				{
					gadget = iihd->ActiveSysGadget;
					iihd->ActiveSysGadget = NULL;

					if (IS_BOOPSI_GADGET(gadget))
					{
						Locked_DoMethodA(window, gadget, (Msg)&gpgi, IntuitionBase);
					}
				}
			}
			break;

		case GTYP_STRGADGET:
		case GTYP_BOOLGADGET:
			gadget->Flags &= ~GFLG_SELECTED;
			break;
		}

		gadget->Activation &= ~GACT_ACTIVEGADGET;

		iihd->ActiveGadget = NULL;
	}

	/* Need this in case of a window created under the input.device task context */
	screen = window->WScreen;

	/* Check if there are still some requesters */
	//jDc: do NOT use this! if there is a requester without ACTIVE flag set this routine will buysloop!
	//jDc: moved this stuff down to layer dispose stuff!
	/*	  while (window->FirstRequest)
		EndRequest(window->FirstRequest, window);*/

	if (!ILOCKCHECK(((struct IntuiActionMsg *)msg))) lock = LockIBase (0);

	if (window == IntuitionBase->ActiveWindow)
	{
		screen->Title = screen->DefaultTitle;
		IntuitionBase->ActiveWindow = NULL;
	}

	if (window == IBase->ActiveToolbarWindow)
		IBase->ActiveToolbarWindow = NULL;

	if (window == iihd->NewActWindow) iihd->NewActWindow = NULL;
	if (window == iihd->LastHelpWindow) iihd->LastHelpWindow = NULL;
	if (IBase->MenuVerifyScreen == IS(screen) &&
		IS(screen)->MenuVerifyActiveWindow == window)
	{
		IBase->MenuVerifyScreen = NULL;
		IS(screen)->MenuVerifyActiveWindow = NULL;
	}
	
	if (IS(screen)->TitlebarBufferWin == window) IS(screen)->TitlebarBufferWin = NULL;

	if (window == HOVERINGWINDOW)
	{
		HOVERINGWINDOW = NULL;
	}

	/* Remove window from the parent/descendant chain and find next active window
	**
	**
	** before:  parent win xyz
	**              \
	**		     \
	**		 deadwindow window
	**		    /
	**		   /
	**		  /
	**	    descendant win abc
	**
	** after: parent win xyz
	**            |
	**            |
	**            |
	**        descendant win abc
	**
	*/

	if (window == window->WScreen->FirstWindow)
	{
		window->WScreen->FirstWindow = window->NextWindow;
	}
	else if ((win2 = window->WScreen->FirstWindow))
	{
		while (win2->NextWindow)
		{
			if (win2->NextWindow == window)
			{
				win2->NextWindow = window->NextWindow;
				break;
			}
			win2 = win2->NextWindow;
		}
	}

	if (window->Descendant)
	{
		window->Descendant->Parent = window->Parent;
	}

	if (window->Parent)
	{
		window->Parent->Descendant = window->Descendant;

	}

	if (!ILOCKCHECK(((struct IntuiActionMsg *)msg))) UnlockIBase (lock);

	// check if this isn't a screenmanager's screen
	#ifdef SKINS
	if (IBase->ScreenManagerPort &&
		(screen->FirstWindow == NULL) && (!(window->MoreFlags & WMFLG_DO_UNLOCKPUBSCREEN)))
	{
		int_SMAsyncAction(SMA_CLOSEWINDOWNOTIFY,(ULONG*)screen,IntuitionBase);
	}
	#endif

	/* Free resources */

	LOCK_REFRESH(screen);

	RenderScreenBar(screen, FALSE, IntuitionBase);

	/* IFont may be NULL if we are called from an OpenWindow failure */
	if (window->IFont)
		CloseFont (window->IFont);

	#ifdef DAMAGECACHE
	if (IW(window)->trashregion) DisposeRegion(IW(window)->trashregion);
	#endif

	if (window->FirstRequest)
	{
		struct Requester *r = window->FirstRequest;
		while (r)
		{
			if (r->ReqLayer) DeleteLayer(0,r->ReqLayer);
			r->ReqLayer = 0;
			r = r->OlderRequest;
		}
	}

	// remove transparency!
#ifdef SKINS
	if (0 == (window->MoreFlags & WMFLG_LAYERMANAGER))
	{
		if (WLAYER(window))
		{
			InstallTransparentRegion(WLAYER(window),NULL);
		}

		if (((struct IntWindow *)(window))->transpregion) DisposeRegion(((struct IntWindow *)(window))->transpregion);
		if (((struct IntWindow *)(window))->gzztranspregion) DisposeRegion(((struct IntWindow *)(window))->gzztranspregion);
	}

	if (IS(window->WScreen)->EventOwnerWindow == window)
	{
		IS(window->WScreen)->EventOwnerWindow = NULL;
	}
#endif

	/* Let the driver clean up. Driver wil dealloc window's rastport */
	intui_CloseWindow (window, IntuitionBase);

	if (!ILOCKCHECK(((struct IntuiActionMsg *)msg))) lock = LockIBase (0);

	if (!IntuitionBase->ActiveWindow)
	{
		DeactivateWindow(window,msg->msg.task, TRUE, IntuitionBase);
	}

	// yes, this was already done above, but let's repeat this in case DeactivateWindow did something
	// really odd or any other odd crap happened
	if (window == IntuitionBase->ActiveWindow) IntuitionBase->ActiveWindow = NULL;
	if (window == IBase->ActiveToolbarWindow)
		IBase->ActiveToolbarWindow = NULL;

	if (window == iihd->NewActWindow) iihd->NewActWindow = NULL;
	if (window == iihd->LastHelpWindow) iihd->LastHelpWindow = NULL;
	if (IBase->MenuVerifyScreen == IS(screen) &&
		IS(screen)->MenuVerifyActiveWindow == window)
	{
		IBase->MenuVerifyScreen = NULL;
		IS(screen)->MenuVerifyActiveWindow = NULL;
	}

	if (IS(screen)->TitlebarBufferWin == window) IS(screen)->TitlebarBufferWin = NULL;

	if (window == HOVERINGWINDOW)
	{
		HOVERINGWINDOW = NULL;
	}

	if (!ILOCKCHECK(((struct IntuiActionMsg *)msg))) UnlockIBase (lock);

	CheckLayers(screen, IntuitionBase);

	UNLOCK_REFRESH(screen);

	if (WMFLG_LAYERMANAGER == (window->MoreFlags & WMFLG_LAYERMANAGER))
	{
		windowLayersClosed(window, IntuitionBase);
	}

	/* jDc: trash the screen pointer to avoid unnecessary checks in WindowValid() and
	   memory corruption */
	window->WScreen = (struct Screen *)0x0;

	/* Free memory for the window */
#ifdef __MORPHOS__
	DisposeObject((Object*)window);
#else
	FreeMem (window, sizeof(struct IntWindow));
#endif


} /* int_closewindow */


/**********************************************************************************/


void intui_CloseWindow (struct Window * w,
						struct IntuitionBase * IntuitionBase)
{
	KillWinSysGadgets(w, IntuitionBase);

	if (0 == (w->MoreFlags & WMFLG_LAYERMANAGER))
	{
		/* a GZZ window */
		/* delete inner window */
		if (NULL != WLAYER(w))
			DeleteLayer(0, WLAYER(w));

		/* delete outer window */
		if (NULL != BLAYER(w))
			DeleteLayer(0, BLAYER(w));
	}
	else
	{
		// Set this early or CheckMouseZone will still find this layer
		LayersControlTags(&w->WScreen->LayerInfo,
			LT_Layer, (ULONG)WLAYER(w),
			LT_HitLevel,255,
			LT_Window, 0,
			LT_Layer, (ULONG)BLAYER(w),
			LT_HitLevel, 255,
			LT_Window, 0,
			TAG_DONE);
	}
	
	if (0 == (w->Flags & WFLG_GIMMEZEROZERO))
	{
		DeinitRastPort(w->BorderRPort);
		FreeMem(w->BorderRPort, sizeof(struct RastPort));
	}

	#ifdef SKINS
	CheckMouseZone(FALSE,IntuitionBase);
	UpdateScreenPointer(w->WScreen,IntuitionBase);
	#endif

	if (IW(w)->free_pointer)
		DisposeObject(IW(w)->pointer);
}
